#include <fcntl.h>
#include <a.out.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/sysinfo.h>
#include <sys/sysmacros.h>
#include <sys/immu.h>
#include <sys/utsname.h>
#include <errno.h>
#include <curses.h>
#include "consts.h"

#define CPU_INFO_SZ (sizeof(sysinfo.cpu) / sizeof(sysinfo.cpu[0]))
#define DEFAULT_HZ "60"

#define GRPH_LINES 9
#define GRPH_COLS 20
#define MAP_LINES (GRPH_LINES - 2)
#define MAP_COLS (GRPH_COLS - 1)

#define DEFAULT_REFRESH 5
#define HEADER "Monitor"

extern long lseek();
extern char *getenv();

extern int GreenBlackPair;
extern int RedBlackPair;


struct sysinfo sysinfo;  /* need to declare these to deal with sys/sysinfo.h */
time_t lbolt;            /* extern declarations */
struct minfo minfo;
int nswap;
int maxmem;
int freemem;

static struct nlist NlistInfo[] = {
#define NL_SYSINFO 0
	{"sysinfo"},
#define NL_LBOLT 1
	{"lbolt"},
#define NL_MINFO 2
	{"minfo"},
#define NL_NSWAP 3
	{"nswap"},
#define NL_FREEMEM 4
	{"freemem"},
#define NL_MAXMEM 5
	{"maxmem"},
	{0}};

typedef WINDOW * display_t;
typedef struct cpu_stats_t {
	double user;
	double kernel;
	double idle;
	} cpu_stats_t;
typedef struct grph_win_t {
	WINDOW *back;
	WINDOW *top;
	int col;
	} grph_win_t;

#ifdef DEBUG
#define sleep(n) {time_t d = time((time_t *)0) + n ; while\
    (d >= time((time_t *)0));}
#endif

static int NlistLoaded = 0;
static int KmemFd;
static time_t SavedCpu[CPU_INFO_SZ];
static time_t BootTime;
static int Hertz;
static time_t Now;
static int OptNoColor = 0;

static void DisplayAdd();
static void InitGraphWin();
static void UpdateGraph();
static void XtraInfoFn();

int main(argc, argv)

int argc;
char *argv[];

	{
	char head[sizeof(HEADER) + sizeof(struct utsname) + 32];
	struct utsname name;
	grph_win_t user_win, kernel_win, swap_win, mem_win;
	cpu_stats_t cpu_stats;
	double free_swap;
	double free_memory;

	if (uname(&name) == -1)
		{
		perror(argv[0]);
		exit(1);
		}
	initscr();
	waddstr(stdscr, "initialising...");
	wrefresh(stdscr);
	if (start_color() != OK)
		OptNoColor = 1;
	else
		if (! has_colors())
			OptNoColor = 1;
		else
			ColorMapInit();
	Now = time((time_t *)0);
	if (! NlistFnInit(stdscr))
		exit(1);
	wmove(stdscr, 0, 0);
	wclrtoeol(stdscr);
	mvwaddstr(stdscr, 0, 0, "wait...");
	wrefresh(stdscr);
	InitGraphWin(&user_win, 2, 0, "User CPU %");
	InitGraphWin(&kernel_win, 2, GRPH_COLS + 2, "Kernel CPU %");
	InitGraphWin(&swap_win, 2, (GRPH_COLS + 2) * 2, "Free swap pages %");
	InitGraphWin(&mem_win, 2 + GRPH_LINES + 1, 0, "Free memory %");
	sleep(2);
	curs_set(FALSE);
	werase(stdscr);
	sprintf(head, "%s - %s %s Unix %s.%s", HEADER, name.sysname,
	    name.machine, name.release, name.version);
	wstandout(stdscr);
	mvwaddstr(stdscr, 0, (COLS - strlen(head)) / 2 - 1, head);
	wstandend(stdscr);
	wrefresh(stdscr);
	wrefresh(user_win.back);
	wrefresh(kernel_win.back);
	wrefresh(swap_win.back);
	wrefresh(mem_win.back);
	for (;;)
		{
		Now = time((time_t *)0);
		(void)NlistFn(stdscr, &cpu_stats);
		XtraInfoFn(stdscr, &free_swap, &free_memory);
		UpdateGraph(&user_win, cpu_stats.user, FIELD_HOT);
		UpdateGraph(&kernel_win, cpu_stats.kernel, FIELD_HOT);
		UpdateGraph(&swap_win, free_swap, FIELD_COLD);
		UpdateGraph(&mem_win, free_memory, FIELD_COLD);
		wmove(stdscr, LINES - 1, 0);
		wrefresh(stdscr);
		sleep(DEFAULT_REFRESH);
		}
	endwin();
	}

static void XtraInfoFn(win, free_swap, free_memory)

WINDOW *win;
double *free_swap;
double *free_memory;

	{
	if (lseek(KmemFd, NlistInfo[NL_MINFO].n_value, 0) == -1L)
		{
		DisplayAdd(win, "kmem lseek error ");
		flash();
		return;
		}
	if (read(KmemFd, (char *)&minfo, sizeof(minfo)) < 0)
		{
		DisplayAdd(win, "kmem read error ");
		flash();
		return;
		}
	if (lseek(KmemFd, NlistInfo[NL_FREEMEM].n_value, 0) == -1L)
		{
		DisplayAdd(win, "kmem lseek error ");
		flash();
		return;
		}
	if (read(KmemFd, (char *)&freemem, sizeof(freemem)) < 0)
		{
		DisplayAdd(win, "kmem read error ");
		flash();
		return;
		}
	wmove(win, LINES - 3, 0);
	wprintw(win, "free swap: %ld nswap: %ld maxmem: %ld freemem: %ld      ",
	    minfo.freeswap, nswap, ctob(maxmem) / NBPP, freemem);
	*free_swap = minfo.freeswap / (double)nswap;
	*free_memory = freemem / (double)(ctob(maxmem) / NBPP);
	wrefresh(win);
	}

static void UpdateGraph(grph_win, level, field)

grph_win_t *grph_win;
double level;
int field;

	{
	int i;
	int spot_y, spot_x;
	double scale;

	spot_y = (double)MAP_LINES - (level * (double)MAP_LINES)
	    + 0.5;
	if (spot_y >= MAP_LINES)
		spot_y = MAP_LINES - 1;
	else
		if (spot_y < 0)
			spot_y = 0;
	spot_x = grph_win->col;
	if (spot_x >= MAP_COLS)
		{
		spot_x = MAP_COLS - 1;
		for (i = 0; i < MAP_LINES; ++i)
			mvwdelch(grph_win->top, i, 0);
		}
	grph_win->col = spot_x + 1;
	if (OptNoColor)
		{
		wstandout(grph_win->top);
		for (i = spot_y; i < MAP_LINES; ++i)
			mvwaddch(grph_win->top, i, spot_x, ' ');
		wstandend(grph_win->top);
		}
	else
		for (i = spot_y; i < MAP_LINES; ++i)
			{
			scale = 1.0 - i / (double)(MAP_LINES - 1);
			FillColor(grph_win->top, scale, i, spot_x, field);
			}
	wrefresh(grph_win->top);
	}

int NlistFnInit(display)

display_t *display;

	{
	int i;
	char *hz_str;
	struct sysinfo sysinfo;
	static int kmem_status = 0;

	if (! NlistLoaded)
		{
		NlistLoaded = 1;
		if (nlist("/unix", NlistInfo) < 0)
			{
			DisplayAdd(display, "nlist error ");
			return 0;
			}
		if (NlistInfo[NL_SYSINFO].n_value == 0)
			{
			DisplayAdd(display, "bad sysinfo offset ");
			flash();
			return 0;
			}
		if (NlistInfo[NL_LBOLT].n_value == 0)
			{
			DisplayAdd(display, "bad lbolt offset ");
			flash();
			return 0;
			}
		if (! (hz_str = getenv("HZ")))
			hz_str = DEFAULT_HZ;
		Hertz = atoi(hz_str);
		if ((KmemFd = open("/dev/kmem", O_RDONLY)) < 0)
			{
			DisplayAdd(display, "cannot open kmem ");
			return 0;
			}
		if (lseek(KmemFd, NlistInfo[NL_SYSINFO].n_value, 0) == -1L)
			{
			char buf[128];
			extern char *sys_errlist[];

			sprintf(buf, "kmem lseek error %d %s %u ", errno,
			    sys_errlist[errno], NlistInfo[NL_SYSINFO].n_value);
			DisplayAdd(display, buf);
			flash();
			return 0;
			}
		if (read(KmemFd, (char *)&sysinfo, sizeof(sysinfo)) < 0)
			{
			DisplayAdd(display, "kmem read error ");
			flash();
			return 0;
			}
		for (i = 0; i < CPU_INFO_SZ; ++i)
			SavedCpu[i] = sysinfo.cpu[i];
		if (lseek(KmemFd, NlistInfo[NL_MINFO].n_value, 0) == -1L)
			{
			char buf[128];
			extern char *sys_errlist[];

			sprintf(buf, "kmem lseek error %d %s %u ", errno,
			    sys_errlist[errno], NlistInfo[NL_MINFO].n_value);
			DisplayAdd(display, buf);
			flash();
			return 0;
			}
		if (read(KmemFd, (char *)&minfo, sizeof(minfo)) < 0)
			{
			DisplayAdd(display, "kmem read error ");
			flash();
			return 0;
			}
		if (lseek(KmemFd, NlistInfo[NL_NSWAP].n_value, 0) == -1L)
			{
			char buf[128];
			extern char *sys_errlist[];

			sprintf(buf, "kmem lseek error %d %s %u ", errno,
			    sys_errlist[errno], NlistInfo[NL_NSWAP].n_value);
			DisplayAdd(display, buf);
			flash();
			return 0;
			}
		if (read(KmemFd, (char *)&nswap, sizeof(nswap)) < 0)
			{
			DisplayAdd(display, "kmem read error ");
			flash();
			return 0;
			}
		if (lseek(KmemFd, NlistInfo[NL_MAXMEM].n_value, 0) == -1L)
			{
			char buf[128];
			extern char *sys_errlist[];

			sprintf(buf, "kmem lseek error %d %s %u ", errno,
			    sys_errlist[errno], NlistInfo[NL_MAXMEM].n_value);
			DisplayAdd(display, buf);
			flash();
			return 0;
			}
		if (read(KmemFd, (char *)&maxmem, sizeof(maxmem)) < 0)
			{
			DisplayAdd(display, "kmem read error ");
			flash();
			return 0;
			}

		kmem_status = 1;
		}

	return kmem_status;
	}

int NlistFn(display, cpu_stats)

display_t *display;
cpu_stats_t *cpu_stats;

	{
	int i;
	time_t all_krn;
	struct sysinfo sysinfo;
	time_t new_cpu[CPU_INFO_SZ];
	double total_cpu;

	if (lseek(KmemFd, NlistInfo[NL_SYSINFO].n_value, 0) == -1L)
		{
		DisplayAdd(display, "kmem lseek error ");
		flash();
		return 0;
		}
	if (read(KmemFd, (char *)&sysinfo, sizeof(sysinfo)) < 0)
		{
		DisplayAdd(display, "kmem read error ");
		flash();
		return 0;
		}

	total_cpu = 0.0;
	for (i = 0; i < CPU_INFO_SZ; ++i)
		{
		total_cpu += new_cpu[i] = sysinfo.cpu[i] - SavedCpu[i];
		SavedCpu[i] = sysinfo.cpu[i];
		}

	cpu_stats->user = new_cpu[CPU_USER] / total_cpu;
#ifndef PARTIAL
	all_krn = new_cpu[CPU_KERNEL] + new_cpu[CPU_WAIT] + new_cpu[CPU_SXBRK];
#else
	all_krn = new_cpu[CPU_KERNEL];
#endif
	cpu_stats->kernel = all_krn / total_cpu;
	cpu_stats->idle = new_cpu[CPU_IDLE] / total_cpu;
	wmove(stdscr, LINES - 2, 0);
	wprintw(stdscr, "swapin: %ld", sysinfo.swapin);
	wmove(stdscr, LINES - 1, 0);
	wprintw(stdscr, "runque: %ld", sysinfo.runque);
	wrefresh(stdscr);

	return 1;
	}

int UptimeFnInit(display)

display_t *display;

	{
	time_t lbolt;                     /* a contraction of lightning bolt */

	if (NlistFnInit(display) == 0)
		return 0;

	if (lseek(KmemFd, NlistInfo[NL_LBOLT].n_value, 0) == -1L)
		{
		DisplayAdd(display, "kmem lbolt lseek error ");
		flash();
		return 0;
		}
	if (read(KmemFd, (char *)&lbolt , sizeof(lbolt)) < 0)
		{
		DisplayAdd(display, "kmem lbolt read error ");
		flash();
		return 0;
		}

	BootTime = Now - lbolt / Hertz;
	if (BootTime < 0)
		DisplayAdd(display, "very strange boottime ");

	return 1;
	}

int UptimeFn(display)

display_t *display;

	{
	time_t lbolt;
	time_t uptime;
	int days, hours, minutes, seconds;
	char buf[64];

	uptime = Now - BootTime;
	days = uptime / (60 * 60 * 24);
	uptime -= days * (60 * 60 * 24);
	hours = uptime / (60 * 60);
	uptime -= hours * (60 * 60);
	minutes = uptime / 60;
	uptime -= minutes * 60;
	seconds = uptime;
	sprintf(buf, "ut: %d+%.2d:%.2d:%.2d ", days, hours, minutes, seconds);
	DisplayAdd(display, buf);

	return 1;
	}

static void DisplayAdd(display, msg)

WINDOW *display;
char *msg;

	{
	wprintw(stdscr, "%s\n", msg);
	wrefresh(stdscr);
	}

static void InitGraphWin(win, start_y, start_x, label)

grph_win_t *win;
int start_y;
int start_x;
char *label;

	{
	int i, j;
	int len;

	win->back = newwin(GRPH_LINES, GRPH_COLS, start_y, start_x);
	if (! OptNoColor)
		wattron(win->back, COLOR_PAIR(GreenBlackPair));
	for (i = 0; i < MAP_LINES + 1; ++i)
		mvwaddch(win->back, i, 0, ACS_VLINE);
	for (i = 0; i < MAP_COLS + 1; ++i)
		mvwaddch(win->back, MAP_LINES, i, ACS_HLINE);
	mvwaddch(win->back, MAP_LINES, 0, ACS_LLCORNER);
	if (! OptNoColor)
		wattroff(win->back, COLOR_PAIR(GreenBlackPair));
	win->top = newwin(MAP_LINES, MAP_COLS, start_y, start_x + GRPH_COLS
	    - MAP_COLS);
	win->col = 0;
	len = strlen(label);
	if (! OptNoColor)
		wattron(win->back, COLOR_PAIR(RedBlackPair));
	mvwaddstr(win->back, GRPH_LINES - 1, (GRPH_COLS - len) / 2, label);
	if (! OptNoColor)
		wattroff(win->back, COLOR_PAIR(RedBlackPair));
	}
